# -*- coding: utf-8 -*-

"""Top-level package for Cell Maps Generate Hierarchy."""

__author__ = 'Cell Maps team'
__email__ = 'tools@cm4ai.org'
__version__ = '0.2.1'
__repo_url__ = 'https://github.com/idekerlab/cellmaps_generate_hierarchy'
__description__ = 'A tool to generate hierarchies from protein to protein interaction networks'
__computation_name__ = 'Hierarchy'
